# -*- coding: utf-8 -*-

"""
Setup automation for Cloudflare.
"""

import typing as T
import os
import subprocess
import dataclasses
from functools import cached_property

try:
    import boto3
    import botocore.exceptions
except ImportError:  # pragma: no cover
    pass

from .vendor.emoji import Emoji
from .vendor.better_pathlib import temp_cwd

from .logger import logger
from .runtime import IS_CI
from .helpers import print_command

if T.TYPE_CHECKING:  # pragma: no cover
    from .define import PyWf
    from boto3 import Session
    from mypy_boto3_codeartifact.client import CodeArtifactClient


@dataclasses.dataclass
class PyWfCloudflare:  # pragma: no cover
    """
    Namespace class for Cloudflare setup automation.
    """

    @cached_property
    def cloudflare_token(self: "PyWf") -> str:
        if self.path_cloudflare_token_file.exists():
            return self.path_cloudflare_token_file.read_text(encoding="utf-8").strip()
        else:  # pragma: no cover
            message = (
                f"{Emoji.error} Cannot find Cloudflare token file at "
                f"{self.path_cloudflare_token_file}!\n"
                f"{self.__class__.path_cloudflare_token_file.__doc__}"
            )
            raise FileNotFoundError(message)

    @logger.emoji_block(
        msg="Create Cloudflare Pages project",
        emoji=Emoji.doc,
    )
    def create_cloudflare_pages_project(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = False,
    ):
        os.environ["CLOUDFLARE_API_TOKEN"] = self.cloudflare_token
        args = [
            f"{self.path_bin_wrangler}",
            "pages",
            "project",
            "create",
            self.package_name_slug,
            "--production-branch",
            "main",
        ]
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)

    @logger.emoji_block(
        msg="Create Cloudflare Pages project",
        emoji=Emoji.doc,
    )
    def deploy_cloudflare_pages(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = False,
    ):
        os.environ["CLOUDFLARE_API_TOKEN"] = self.cloudflare_token
        args = [
            f"{self.path_bin_wrangler}",
            "pages",
            "deploy",
            f"{self.dir_sphinx_doc_build_html}",
            f"--project-name={self.package_name_slug}",
        ]
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)



# -*- coding: utf-8 -*-

"""
Document Build and Deploy Automation for Python Projects.
"""

import typing as T
import shutil
import dataclasses
import subprocess

from .vendor.emoji import Emoji
from .vendor.os_platform import OPEN_COMMAND

from .logger import logger
from .helpers import print_command

if T.TYPE_CHECKING:  # pragma: no cover
    from .define import PyWf


@dataclasses.dataclass
class PyWfDocs:
    """
    Namespace class for document related automation.
    """

    @logger.emoji_block(
        msg="Build Documentation Site Locally",
        emoji=Emoji.doc,
    )
    def _build_doc(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Use sphinx doc to build documentation site locally. It set the
        necessary environment variables so that the ``make html`` command
        can build the HTML successfully.

        Run:

        .. code-block:: bash

            sphinx-build -M html docs/source docs/build
        """
        if real_run:
            shutil.rmtree(
                f"{self.dir_sphinx_doc_build}",
                ignore_errors=True,
            )
            shutil.rmtree(
                f"{self.dir_sphinx_doc_source_python_lib}",
                ignore_errors=True,
            )

        args = [
            f"{self.path_venv_bin_sphinx_build}",
            "-M",
            "html",
            f"{self.dir_sphinx_doc_source}",
            f"{self.dir_sphinx_doc_build}",
        ]
        print_command(args)
        if real_run:
            subprocess.run(args, check=True)

    def build_doc(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(not verbose):
            return self._build_doc(
                real_run=real_run,
                quiet=not verbose,
            )

    build_doc.__doc__ = _build_doc.__doc__

    @logger.emoji_block(
        msg="View Documentation Site Locally",
        emoji=Emoji.doc,
    )
    def _view_doc(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        View documentation site built locally in web browser.

        It is usually at the ``${dir_project_root}/build/html/index.html``

        Run:

        .. code-block:: bash

            # For MacOS / Linux
            open build/html/index.html
            # For Windows
            start build/html/index.html
        """
        args = [OPEN_COMMAND, f"{self.path_sphinx_doc_build_index_html}"]
        print_command(args)
        if real_run:
            subprocess.run(args)

    def view_doc(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(not verbose):
            return self._view_doc(
                real_run=real_run,
                quiet=not verbose,
            )

    view_doc.__doc__ = _view_doc.__doc__

    @logger.emoji_block(
        msg="Deploy Documentation Site To S3 as Versioned Doc",
        emoji=Emoji.doc,
    )
    def _deploy_versioned_doc(
        self: "PyWf",
        bucket: T.Optional[str] = None,
        prefix: str = "projects/",
        aws_profile: T.Optional[str] = None,
        real_run: bool = True,
        quiet: bool = False,
    ) -> bool:  # pragma: no cover
        """
        Deploy versioned document to AWS S3.

        The S3 bucket has to enable static website hosting. The document site
        will be uploaded to ``s3://${bucket}/${prefix}${package_name}/${package_version}/``
        """
        if bool(self.doc_host_s3_bucket) is False:
            logger.info(f"{Emoji.red_circle} doc_host_s3_bucket is not set, skip.")
            return False
        if bucket is None:
            bucket = self.doc_host_s3_bucket
        if aws_profile is None:
            aws_profile = self.doc_host_aws_profile
        args = [
            f"{self.path_bin_aws}",
            "s3",
            "sync",
            f"{self.dir_sphinx_doc_build_html}",
            f"s3://{bucket}/{prefix}{self.package_name}/{self.package_version}/",
        ]
        if aws_profile:
            args.extend(["--profile", aws_profile])
        print_command(args)
        if real_run:
            subprocess.run(args, check=True)
        return real_run

    def deploy_versioned_doc(
        self: "PyWf",
        bucket: T.Optional[str] = None,
        prefix: str = "projects/",
        aws_profile: T.Optional[str] = None,
        real_run: bool = True,
        verbose: bool = True,
    ) -> bool:  # pragma: no cover
        with logger.disabled(not verbose):
            return self._deploy_versioned_doc(
                bucket=bucket,
                prefix=prefix,
                aws_profile=aws_profile,
                real_run=real_run,
                quiet=not verbose,
            )

    deploy_versioned_doc.__doc__ = _deploy_versioned_doc.__doc__

    @logger.emoji_block(
        msg="Deploy Documentation Site To S3 as Latest Doc",
        emoji=Emoji.doc,
    )
    def _deploy_latest_doc(
        self: "PyWf",
        bucket: T.Optional[str] = None,
        prefix: str = "projects/",
        aws_profile: T.Optional[str] = None,
        real_run: bool = True,
        quiet: bool = False,
    ) -> bool:  # pragma: no cover
        """
        Deploy the latest document to AWS S3.

        The S3 bucket has to enable static website hosting. The document site
        will be uploaded to ``s3://${bucket}/${prefix}${package_name}/latest/``
        """
        if bool(self.doc_host_s3_bucket) is False:
            logger.info(f"{Emoji.red_circle} doc_host_s3_bucket is not set, skip.")
            return False
        if bucket is None:
            bucket = self.doc_host_s3_bucket
        if aws_profile is None:
            aws_profile = self.doc_host_aws_profile
        args = [
            f"{self.path_bin_aws}",
            "s3",
            "sync",
            f"{self.dir_sphinx_doc_build_html}",
            f"s3://{bucket}/{prefix}{self.package_name}/latest/",
        ]
        if aws_profile:
            args.extend(["--profile", aws_profile])
        print_command(args)
        if real_run:
            subprocess.run(args, check=True)
        return real_run

    def deploy_latest_doc(
        self: "PyWf",
        bucket: T.Optional[str] = None,
        prefix: str = "projects/",
        aws_profile: T.Optional[str] = None,
        real_run: bool = True,
        verbose: bool = True,
    ) -> bool:  # pragma: no cover
        with logger.disabled(not verbose):
            return self._deploy_latest_doc(
                bucket=bucket,
                prefix=prefix,
                aws_profile=aws_profile,
                real_run=real_run,
                quiet=not verbose,
            )

    deploy_latest_doc.__doc__ = _deploy_latest_doc.__doc__

    @logger.emoji_block(
        msg="View Latest Doc on AWS S3",
        emoji=Emoji.doc,
    )
    def _view_latest_doc(
        self: "PyWf",
        bucket: T.Optional[str] = None,
        prefix: str = "projects/",
        real_run: bool = True,
        quiet: bool = False,
    ):  # pragma: no cover
        """
        Open the latest document that hosted on AWS S3 in web browser.

        Here's a sample document site url
        https://my-bucket.s3.amazonaws.com/my-prefix/my_package/latest/index.html
        """
        if bool(self.doc_host_s3_bucket) is False:
            logger.info(f"{Emoji.red_circle} doc_host_s3_bucket is not set, skip.")
            return False
        if bucket is None:
            bucket = self.doc_host_s3_bucket
        url = (
            f"https://{bucket}.s3.amazonaws.com/{prefix}{self.package_name}"
            f"/latest/{self.path_sphinx_doc_build_index_html.name}"
        )
        args = [OPEN_COMMAND, url]
        print_command(args)
        if real_run:
            subprocess.run(args, check=True)

    def view_latest_doc(
        self: "PyWf",
        bucket: T.Optional[str] = None,
        prefix: str = "projects/",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(not verbose):
            return self._view_latest_doc(
                bucket=bucket,
                prefix=prefix,
                real_run=real_run,
                quiet=not verbose,
            )

    view_latest_doc.__doc__ = _view_latest_doc.__doc__


# -*- coding: utf-8 -*-

"""
Helper functions.

.. note::

    This module is "ZERO-DEPENDENCY".
"""

import typing as T
import hashlib

try:
    from requests import Response
except ImportError as e:  # pragma: no cover
    pass

from .logger import logger


def sha256_of_bytes(b: bytes) -> str:  # pragma: no cover
    sha256 = hashlib.sha256()
    sha256.update(b)
    return sha256.hexdigest()


def extract_digit_tokens(text: str) -> T.List[str]:
    """
    Extract all consecutive digit tokens from a string.

    Example:

        >>> extract_digit_tokens("1.23.456")
        ["1", "23", "456"]
    """
    for c in list(text):
        if not c.isdigit():
            text = text.replace(c, " ")
    return [token.strip() for token in text.split() if token.strip()]


def bump_version(
    current_version: str,
    major: bool = False,
    minor: bool = False,
    patch: bool = False,
    minor_start_from: int = 0,
    micro_start_from: int = 0,
):
    """
    Bump a semantic version. The current version has to be in x.y.z format,
    where x, y, z are integers.

    :param current_version: current version string.
    :param major: bump major version.
    :param minor: bump minor version.
    :param patch: bump patch version.
    :param minor_start_from: if bumping major version, minor start from this number.
    :param micro_start_from: if bumping minor version, micro start from this number.
    """
    if sum([patch, minor, major]) != 1:
        raise ValueError(
            "Only one and exact one of 'patch', 'minor', 'major' can be True"
        )

    # get the current version
    major_ver, minor_ver, micro_ver = current_version.split(".")
    major_ver, minor_ver, micro_ver = int(major_ver), int(minor_ver), int(micro_ver)

    # update version
    if major:
        major_ver += 1
        minor_ver, micro_ver = minor_start_from, micro_start_from
    elif minor:
        minor_ver += 1
        micro_ver = micro_start_from
    elif patch:
        micro_ver += 1
    else:  # pragma: no cover
        raise NotImplementedError

    return f"{major_ver}.{minor_ver}.{micro_ver}"


def print_command(args: T.List[str]):
    cmd = " ".join(args)
    logger.info(f"run command: {cmd}")


def raise_http_response_error(response: "Response"):  # pragma: no cover
    print(f"status = {response.status_code}")
    print(f"body = {response.text}")
    raise Exception("HTTP request failed, see error details above.")



# -*- coding: utf-8 -*-

"""
Testing Automation for Python Projects.
"""

import typing as T
import subprocess
import dataclasses

from .vendor.emoji import Emoji
from .vendor.os_platform import OPEN_COMMAND
from .vendor.better_pathlib import temp_cwd

from .logger import logger
from .helpers import print_command

if T.TYPE_CHECKING:  # pragma: no cover
    from .define import PyWf


@dataclasses.dataclass
class PyWfTests:  # pragma: no cover
    """
    Namespace class for testing related automation.
    """

    @logger.emoji_block(
        msg="Run Unit Test",
        emoji=Emoji.test,
    )
    def _run_unit_test(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        A wrapper of ``pytest`` command to run unit test.

        Run:

        .. code-block:: bash

            pytest tests -s --rootdir=/path/to/project/root
        """
        args = [
            f"{self.path_venv_bin_pytest}",
            f"{self.dir_tests}",
            "-s",
            f"--rootdir={self.dir_project_root}",
        ]
        if quiet:
            args.append("--quiet")
        print_command(args)
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)

    def run_unit_test(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(not verbose):
            return self._run_unit_test(
                real_run=real_run,
                quiet=not verbose,
            )

    run_unit_test.__doc__ = _run_unit_test.__doc__

    @logger.emoji_block(
        msg="Run Code Coverage Test",
        emoji=Emoji.test,
    )
    def _run_cov_test(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        A wrapper of ``pytest`` command to run code coverage test.

        Run:

        .. code-block:: bash

            pytest -s --tb=native --rootdir=/path/to/project/root --cov=package_name --cov-report term-missing --cov-report html:/path/to/htmlcov tests
        """
        args = [
            f"{self.path_venv_bin_pytest}",
            "-s",
            "--tb=native",
            f"--rootdir={self.dir_project_root}",
            f"--cov={self.package_name}",
            "--cov-report",
            "term-missing",
            "--cov-report",
            f"html:{self.dir_htmlcov}",
            f"{self.dir_tests}",
        ]
        if quiet:
            args.append("--quiet")
        print_command(args)
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)

    def run_cov_test(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(not verbose):
            return self._run_cov_test(
                real_run=real_run,
                quiet=not verbose,
            )

    run_cov_test.__doc__ = _run_cov_test.__doc__

    @logger.emoji_block(
        msg="View Code Coverage Test Result",
        emoji=Emoji.test,
    )
    def _view_cov(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        View coverage test output html file locally in web browser.

        It is usually at the ``${dir_project_root}/htmlcov/index.html``

        .. code-block:: bash

            # For MacOS / Linux
            open htmlcov/index.html
            # For Windows
            start htmlcov/index.html
        """
        args = [OPEN_COMMAND, f"{self.path_htmlcov_index_html}"]
        if real_run:
            subprocess.run(args)

    def view_cov(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(not verbose):
            return self._view_cov(
                real_run=real_run,
                quiet=not verbose,
            )

    view_cov.__doc__ = _view_cov.__doc__

    @logger.emoji_block(
        msg="Run Integration Tests",
        emoji=Emoji.test,
    )
    def _run_int_test(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        A wrapper of ``pytest`` command to run integration test.

        Run:

        .. code-block:: bash

            pytest tests_int -s --rootdir=/path/to/project/root
        """
        args = [
            f"{self.path_venv_bin_pytest}",
            f"{self.dir_tests_int}",
            "-s",
            f"--rootdir={self.dir_project_root}",
        ]
        if quiet:
            args.append("--quiet")
        print_command(args)
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)

    def run_int_test(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(not verbose):
            return self._run_int_test(
                real_run=real_run,
                quiet=not verbose,
            )

    run_int_test.__doc__ = _run_int_test.__doc__

    @logger.emoji_block(
        msg="Run Load Test",
        emoji=Emoji.test,
    )
    def _run_load_test(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        A wrapper of ``pytest`` command to run load test.

        Run:

        .. code-block:: bash

            pytest tests_load -s --rootdir=/path/to/project/root
        """
        args = [
            f"{self.path_venv_bin_pytest}",
            f"{self.dir_tests_load}",
            "-s",
            f"--rootdir={self.dir_project_root}",
        ]
        if quiet:
            args.append("--quiet")
        print_command(args)
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)

    def run_load_test(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(not verbose):
            return self._run_load_test(
                real_run=real_run,
                quiet=not verbose,
            )

    run_load_test.__doc__ = _run_load_test.__doc__


# -*- coding: utf-8 -*-

"""
Enumerate useful, UTF8 emoji characters.

Full list is here: https://unicode.org/emoji/charts/full-emoji-list.html

Usage example:

.. code-block:: python

    from emoji import Emoji
"""

__version__ = "0.1.1"


class Emoji:
    start_timer = "⏱"
    end_timer = "⏰"
    start = "⏯"
    end = "⏹"
    error = "🔥"

    relax = "🌴"

    doc = "📔"
    test = "🧪"
    install = "💾"
    build = "🪜"
    deploy = "🚀"
    delete = "🗑"
    config = "🔯"
    tada = "🎉"

    cloudformation = "🐑"
    awslambda = "λ"
    s3 = "🪣"

    template = "📋"
    computer = "💻"
    package = "📦"
    email = "📫"
    factory = "🏭"
    no_entry = "🚫"
    warning = "❗"
    label = "🏷"

    thumb_up = "👍"
    thumb_down = "👎"
    attention = "👉"

    happy_face = "😀"
    hot_face = "🥵"
    anger = "💢"
    eye = "👀"

    red_circle = "🔴"
    green_circle = "🟢"
    yellow_circle = "🟡"
    blue_circle = "🔵"

    red_square = "🟥"
    green_square = "🟩"
    yellow_square = "🟨"
    blue_square = "🟦"

    succeeded = "✅"
    failed = "❌"

    arrow_up = "⬆"
    arrow_down = "⬇"
    arrow_left = "⬅"
    arrow_right = "➡"

    python = "🐍"

    sbx = "📦"
    tst = "🧪"
    prd = "🏭"

    install_phase = "🌱"
    pre_build_phase = "🌿"
    build_phase = "🍀"
    post_build_phase = "🌲"


if __name__ == "__main__":
    chars = list()
    for k, v in Emoji.__dict__.items():
        if not k.startswith("_"):
            if len(v) != 1:
                print(f"{k} = {v}, len = {len(v)}")

            # if len(v) == 1:
            chars.append(v)

    print(" ".join(chars))



# -*- coding: utf-8 -*-

"""
Comprehensive Automation Namespace for Python Project Operations

This module provides a unified interface for managing various aspects of a Python project
lifecycle, including virtual environment, dependencies, testing, documentation,
build, and publishing processes.

The PyWf (Python Workflow) class consolidates multiple operational concerns into a
single, cohesive management system, enabling streamlined project automation.

The class uses a composition-based approach, inheriting from multiple specialized
namespace classes to provide a comprehensive project management toolkit.
"""

import typing as T

try:
    import tomllib
except ImportError:
    import toml as tomllib
import dataclasses
from pathlib import Path

from .define_01_paths import PyWfPaths
from .define_02_venv import PyWfVenv
from .define_03_deps import PyWfDeps
from .define_04_tests import PyWfTests
from .define_05_docs import PyWfDocs
from .define_06_build import PyWfBuild
from .define_07_publish import PyWfPublish
from .define_08_saas import PyWfSaas
from .define_09_aws import PyWfAws
from .define_10_cloudflare import PyWfCloudflare


@dataclasses.dataclass
class PyWf(
    PyWfPaths,
    PyWfVenv,
    PyWfDeps,
    PyWfTests,
    PyWfDocs,
    PyWfBuild,
    PyWfPublish,
    PyWfSaas,
    PyWfAws,
    PyWfCloudflare,
):
    """
    Unified Automation Interface for Python Project Management

    :param dir_project_root: Root directory of the project, typically the git
        repository root. It has to have a ``pyproject.toml`` file in it.
    :param toml_data: Parsed configuration data from ``pyproject.toml``
    """

    dir_project_root: Path = dataclasses.field()
    toml_data: T.Dict[str, T.Any] = dataclasses.field()

    # --------------------------------------------------------------------------
    # [project] Configuration Properties
    # --------------------------------------------------------------------------
    @property
    def package_name(self) -> str:
        """Retrieve the package name from pyproject.toml."""
        return self.toml_data["project"]["name"]

    @property
    def package_name_slug(self) -> str:
        """Slugify package name for use in PyPI, readthedocs, etc."""
        return self.package_name.replace("_", "-")

    @property
    def package_version(self) -> str:
        """Retrieve the package version from pyproject.toml."""
        return self.toml_data["project"]["version"]

    @property
    def package_license(self) -> str:
        """Retrieve the package license from pyproject.toml."""
        return self.toml_data["project"]["license"]

    @property
    def package_description(self) -> str:
        """Retrieve the package description from pyproject.toml."""
        return self.toml_data["project"]["description"]

    @property
    def package_author_name(self) -> str:
        """Retrieve the primary author's name from pyproject.toml."""
        return self.toml_data["project"]["authors"][0]["name"]

    @property
    def package_author_email(self) -> str:
        """Retrieve the primary author's email from pyproject.toml."""
        return self.toml_data["project"]["authors"][0]["email"]

    @property
    def package_maintainer_name(self) -> str:
        """Retrieve the primary maintainer's name from pyproject.toml."""
        return self.toml_data["project"]["maintainers"][0]["name"]

    @property
    def package_maintainer_email(self) -> str:
        """Retrieve the primary maintainer's email from pyproject.toml."""
        return self.toml_data["project"]["maintainers"][0]["email"]

    # --------------------------------------------------------------------------
    # [tool.pywf] Configuration Properties
    # --------------------------------------------------------------------------
    @property
    def py_ver_major(self) -> int:
        """Extract major version number from development Python version."""
        return int(self.toml_data["tool"]["pywf"]["dev_python"].split(".")[0])

    @property
    def py_ver_minor(self) -> int:
        """Extract minor version number from development Python version."""
        return int(self.toml_data["tool"]["pywf"]["dev_python"].split(".")[1])

    @property
    def py_ver_micro(self) -> int:
        """Extract micro version number from development Python version."""
        return int(self.toml_data["tool"]["pywf"]["dev_python"].split(".")[2])

    # --- GitHub.com
    @property
    def github_account(self) -> str:
        return self.toml_data["tool"]["pywf"]["github_account"]

    @property
    def github_token_name(self) -> str:
        return self.toml_data["tool"]["pywf"]["github_token_name"]

    @property
    def git_repo_name(self) -> str:
        """
        Git repo name.
        """
        return self.dir_project_root.name

    @property
    def github_repo_fullname(self) -> str:
        return f"{self.github_account}/{self.git_repo_name}"

    @property
    def github_repo_url(self) -> str:
        return f"https://github.com/{self.github_repo_fullname}"

    @property
    def github_actions_secrets_settings_url(self) -> str:
        return f"{self.github_repo_url}/settings/secrets/actions"

    @property
    def github_versioned_release_url(self) -> str:
        return f"{self.github_repo_url}/releases/tag/{self.package_version}"

    # --- codecov.io
    @property
    def codecov_account(self) -> str:
        return self.toml_data["tool"]["pywf"]["codecov_account"]

    @property
    def codecov_token_name(self) -> str:
        return self.toml_data["tool"]["pywf"]["codecov_token_name"]

    # --- AWS
    @property
    def aws_region(self) -> str:
        return self.toml_data["tool"]["pywf"]["aws_region"]

    @property
    def aws_codeartifact_profile(self) -> str:
        return self.toml_data["tool"]["pywf"]["aws_codeartifact_profile"]

    @property
    def aws_codeartifact_domain(self) -> str:
        return self.toml_data["tool"]["pywf"]["aws_codeartifact_domain"]

    @property
    def aws_codeartifact_repository(self) -> str:
        return self.toml_data["tool"]["pywf"]["aws_codeartifact_repository"]

    @property
    def doc_host_aws_profile(self) -> str:  # pragma: no cover
        """Retrieve AWS profile for documentation hosting."""
        return self.toml_data["tool"]["pywf"]["doc_host_aws_profile"]

    @property
    def doc_host_s3_bucket(self) -> str:  # pragma: no cover
        """Retrieve S3 bucket for documentation hosting."""
        return self.toml_data["tool"]["pywf"]["doc_host_s3_bucket"]

    @property
    def doc_host_s3_prefix(self) -> str:  # pragma: no cover
        """
        Retrieve and sanitize S3 prefix for documentation hosting.

        Ensures prefix does not start or end with '/' to maintain
        consistent path formatting.
        """
        doc_host_s3_prefix = self.toml_data["tool"]["pywf"]["doc_host_s3_prefix"]
        if doc_host_s3_prefix.startswith("/"):
            doc_host_s3_prefix = doc_host_s3_prefix[1:]
        if doc_host_s3_prefix.endswith("/"):
            doc_host_s3_prefix = doc_host_s3_prefix[:-1]
        return doc_host_s3_prefix

    @property
    def cloudflare_account_alias(self) -> str:  # pragma: no cover
        """
        CloudFlare account alias, it is an alias you made up for your account
        and we need this to locate the token file.
        """
        return self.toml_data["tool"]["pywf"]["cloudflare_account_alias"]

    @property
    def cloudflare_pages_token_name(self) -> str:  # pragma: no cover
        """
        The Cloudflare API token you use to upload the documentation to Cloudflare Pages.
        """
        return self.toml_data["tool"]["pywf"]["cloudflare_pages_token_name"]

    def _validate_paths(self):
        """
        Validate project root directory and package structure.

        Checks:

        - Verify presence of ``pyproject.toml``
        - Confirm ``package/__init__.py`` exists
        """
        if isinstance(self.dir_project_root, Path) is False:
            self.dir_project_root = Path(self.dir_project_root)

        if self.dir_project_root.joinpath("pyproject.toml").exists() is False:
            raise ValueError(
                f"{self.dir_project_root} does not have a pyproject.toml file! "
                f"it might not be a valid project root directory."
            )
        dir_python_lib = self.dir_project_root.joinpath(self.package_name)
        if dir_python_lib.joinpath("__init__.py").exists() is False:
            raise ValueError(
                f"{dir_python_lib} does not have a __init__.py file, "
                f"the package name {self.package_name} might be invalid."
            )

    def _validate_python_version(self):
        """
        Validate the Python version used in the project.
        """
        if self.py_ver_major != 3:
            raise ValueError(
                f"Python major version has to be 3, but got {self.py_ver_major}."
            )
        if self.py_ver_minor < 11:
            raise ValueError(
                f"PyWf tool only support Python3.11+, but got {self.py_ver_major}.{self.py_ver_minor}"
            )

    def _update_version_file(self):
        """
        Update the version file with current project metadata in ``pyproject.toml``.
        """
        dir_here = Path(__file__).absolute().parent
        path_version_tpl = dir_here / "_version.tpl"
        content = path_version_tpl.read_text(encoding="utf-8").format(
            version=self.package_version,
            description=self.package_description,
            license=self.package_license,
            author=self.package_author_name,
            author_email=self.package_author_email,
            maintainer=self.package_maintainer_name,
            maintainer_email=self.package_maintainer_email,
        )
        self.path_version_py.write_text(content, encoding="utf-8")

    def __post_init__(self):
        self._validate_paths()
        self._validate_python_version()
        self._update_version_file()

    @classmethod
    def from_pyproject_toml(
        cls,
        path_pyproject_toml: Path,
    ):
        """
        Create a :class:`PyWf` instance from a pyproject.toml file.
        """
        path_pyproject_toml = Path(path_pyproject_toml)
        toml_data = tomllib.loads(path_pyproject_toml.read_text(encoding="utf-8"))
        return cls(
            dir_project_root=path_pyproject_toml.parent,
            toml_data=toml_data,
        )



# -*- coding: utf-8 -*-

"""
Setup automation for AWS services.
"""

import typing as T
import os
import subprocess
import dataclasses
from functools import cached_property

try:
    import boto3
    import botocore.exceptions
except ImportError:  # pragma: no cover
    pass

from .vendor.emoji import Emoji
from .vendor.better_pathlib import temp_cwd

from .logger import logger
from .runtime import IS_CI
from .helpers import sha256_of_bytes, print_command

if T.TYPE_CHECKING:  # pragma: no cover
    from .define import PyWf
    from boto3 import Session
    from mypy_boto3_codeartifact.client import CodeArtifactClient


@dataclasses.dataclass
class PyWfAws:  # pragma: no cover
    """
    Namespace class for AWS setup automation.
    """

    @cached_property
    def boto_ses_codeartifact(self: "PyWf") -> "Session":
        if self.aws_codeartifact_profile:
            profile_name = self.aws_codeartifact_profile
        else:
            profile_name = None
        return boto3.Session(
            profile_name=profile_name,
            region_name=self.aws_region,
        )

    @cached_property
    def codeartifact_client(self: "PyWf") -> "CodeArtifactClient":
        return self.boto_ses_codeartifact.client("codeartifact")

    def get_codeartifact_repository_endpoint(self: "PyWf") -> str:
        """
        Reference:

        - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/codeartifact/client/get_repository_endpoint.html
        """
        res = self.codeartifact_client.get_repository_endpoint(
            domain=self.aws_codeartifact_domain,
            repository=self.aws_codeartifact_repository,
            format="pypi",
        )
        return res["repositoryEndpoint"]

    def get_codeartifact_authorization_token(self: "PyWf") -> str:
        """
        Reference:

        - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/codeartifact/client/get_authorization_token.html
        """
        res = self.codeartifact_client.get_authorization_token(
            domain=self.aws_codeartifact_domain,
        )
        return res["authorizationToken"]

    @property
    def poetry_secondary_source_name(self: "PyWf") -> str:
        return self.aws_codeartifact_repository

    @logger.emoji_block(
        msg="Add CodeArtifact as a secondary source",
        emoji=Emoji.install,
    )
    def _poetry_source_add_codeartifact(
        self: "PyWf",
        codeartifact_repository_endpoint: str,
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Run:

        .. code-block:: bash

            poetry source add --secondary ${source_name} "https://${domain_name}-${aws_account_id}.d.codeartifact.${aws_region}.amazonaws.com/pypi/${repository_name}/simple/"
        """
        args = [
            f"{self.path_bin_poetry}",
            "source",
            "add",
            "--priority=supplemental",
            self.poetry_secondary_source_name,
            f"{codeartifact_repository_endpoint}simple/",
        ]
        print_command(args)
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)

    def poetry_source_add_codeartifact(
        self: "PyWf",
        codeartifact_repository_endpoint: str,
        real_run: bool = True,
        verbose: bool = True,
    ):
        with logger.disabled(disable=not verbose):
            self._poetry_source_add_codeartifact(
                codeartifact_repository_endpoint=codeartifact_repository_endpoint,
                real_run=real_run,
                quiet=not verbose,
            )

    poetry_source_add_codeartifact.__doc__ = _poetry_source_add_codeartifact.__doc__

    def poetry_authorization(
        self: "PyWf",
        codeartifact_authorization_token: str,
        real_run: bool = True,
        verbose: bool = True,
    ):
        """
        Set environment variables to allow Poetry to authenticate with CodeArtifact.
        """
        token = codeartifact_authorization_token
        source_name = self.poetry_secondary_source_name.upper()
        if real_run:
            os.environ[f"POETRY_HTTP_BASIC_{source_name}_USERNAME"] = "aws"
            os.environ[f"POETRY_HTTP_BASIC_{source_name}_PASSWORD"] = token

    def twine_authorization(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):
        """
        Run

        .. code-block:: bash

            aws codeartifact login --tool twine \
                --domain ${domain_name} \
                --domain-owner ${aws_account_id} \
                --repository ${repo_name} \
                --profile ${aws_profile}

        Reference:

        - `Configure and use twine with CodeArtifact <https://docs.aws.amazon.com/codeartifact/latest/ug/python-configure-twine.html>`_
        - `AWS CodeArtifact CLI <https://docs.aws.amazon.com/cli/latest/reference/codeartifact/index.html>`_
        """
        aws_account_id = self.boto_ses_codeartifact.client("sts").get_caller_identity()[
            "Account"
        ]
        args = [
            f"{self.path_bin_aws}",
            "codeartifact",
            "login",
            "--tool",
            "twine",
            "--domain",
            self.aws_codeartifact_domain,
            "--domain-owner",
            aws_account_id,
            "--repository",
            self.aws_codeartifact_repository,
        ]
        if IS_CI is False:
            if self.aws_codeartifact_profile:
                args.extend(["--profile", self.aws_codeartifact_profile])

        print_command(args)
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)

    @logger.emoji_block(
        msg="Run twine upload command",
        emoji=Emoji.package,
    )
    def _twine_upload(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Upload Python package to CodeArtifact.

        Run

        .. code-block:: bash

            twine upload dist/* --repository codeartifact
        """
        aws_account_id = self.boto_ses_codeartifact.client("sts").get_caller_identity()[
            "Account"
        ]
        aws_region = self.aws_region
        console_url = (
            f"https://{aws_region}.console.aws.amazon.com/codesuite/codeartifact/d"
            f"/{aws_account_id}/{self.aws_codeartifact_domain}/r"
            f"/{self.aws_codeartifact_repository}/p/pypi/"
            f"{self.package_name}/versions?region={aws_region}"
        )
        logger.info(f"preview in AWS CodeArtifact console: {console_url}")
        args = [
            f"{self.path_bin_twine}",
            "upload",
            f"{self.dir_dist}/*",
            "--repository",
            "codeartifact",
        ]
        print_command(args)
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)

    def twine_upload(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):
        with logger.disabled(disable=not verbose):
            self._twine_upload(
                real_run=real_run,
                quiet=not verbose,
            )

    twine_upload.__doc__ = _twine_upload.__doc__

    @logger.emoji_block(
        msg="Publish Python package to AWS CodeArtifact",
        emoji=Emoji.package,
    )
    def _publish_to_codeartifact(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):
        """
        Publish your Python package to AWS CodeArtifact
        """
        try:
            self.codeartifact_client.describe_package_version(
                domain=self.aws_codeartifact_domain,
                repository=self.aws_codeartifact_repository,
                format="pypi",
                package=self.package_name_slug,
                packageVersion=self.package_version,
            )
            message = (
                f"package {self.package_name_slug!r} "
                f"= {self.package_version} already exists!"
            )
            raise Exception(message)
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                pass
            else:
                raise e

        self.twine_authorization(real_run=real_run, verbose=verbose)
        with logger.nested():
            self.twine_upload(real_run=real_run, verbose=verbose)

    def publish_to_codeartifact(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):
        with logger.disabled(disable=not verbose):
            self._publish_to_codeartifact(
                real_run=real_run,
                verbose=verbose,
            )

    publish_to_codeartifact.__doc__ = _publish_to_codeartifact.__doc__

    @logger.emoji_block(
        msg="Remove package version from AWS CodeArtifact",
        emoji=Emoji.package,
    )
    def remove_from_codeartifact(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):
        """
        Remove a specific version of your Python package release AWS CodeArtifact.

        .. warning::

            I suggest don't do this unless you have to.
            If you re-publish your package with the same version,
            then you may need to invalid the poetry cache before doing poetry lock.
        """
        res = input(
            f"Are you sure you want to remove the package {self.package_name_slug!r} "
            f"version {self.package_version!r}? (Y/N): "
        )
        if res == "Y":
            if real_run:
                self.codeartifact_client.delete_package_versions(
                    domain=self.aws_codeartifact_domain,
                    repository=self.aws_codeartifact_repository,
                    format="pypi",
                    package=self.package_name_slug,
                    versions=[self.package_version],
                    expectedStatus="Published",
                )
                logger.info("Package version removed.")
            else:
                logger.info("Not a real run, do nothing.")
        else:
            logger.info("Aborted")



# -*- coding: utf-8 -*-

"""
Dependency Management Automation for Python Projects.

This module provides comprehensive automation for managing project dependencies
using `Poetry <<https://python-poetry.org/>`_, including locking, installation,
and export ``requirements.txt`` functionalities.
"""

import typing as T
import json
import subprocess
import dataclasses
from pathlib import Path

from .vendor.emoji import Emoji
from .vendor.better_pathlib import temp_cwd

from .logger import logger
from .helpers import sha256_of_bytes, print_command


if T.TYPE_CHECKING:  # pragma: no cover
    from .define import PyWf


@dataclasses.dataclass
class PyWfDeps:
    """
    Namespace class for managing project dependencies using Poetry.
    """

    def _run_poetry_command(
        self: "PyWf",
        args: T.List[str],
        real_run: bool,
        quiet: bool,
    ):
        args = [f"{self.path_bin_poetry}", *args]
        if quiet:
            args.append("--quiet")
        print_command(args)
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)

    @logger.emoji_block(
        msg="Resolve Dependencies Tree",
        emoji=Emoji.install,
    )
    def _poetry_lock(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Lock project dependencies defined in pyproject.toml.

        Resolves and writes exact dependency versions to ``poetry.lock`` file.
        This ensures reproducible builds across different environments.
        You have to run this everytime you changed the ``pyproject.toml`` file.
        And you should commit the latest ``poetry.lock`` file to git.

        Run:

        .. code-block:: bash

            poetry lock

        Ref:

        - poetry lock: https://python-poetry.org/docs/cli/#lock
        """
        return self._run_poetry_command(
            args=["lock"],
            real_run=real_run,
            quiet=quiet,
        )

    def poetry_lock(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(disable=not verbose):
            return self._poetry_lock(
                real_run=real_run,
                quiet=not verbose,
            )

    poetry_lock.__doc__ = _poetry_lock.__doc__

    @logger.emoji_block(
        msg="Install package source code without any dependencies",
        emoji=Emoji.install,
    )
    def _poetry_install_only_root(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Install only the package source code in editable mode.

        Skips installing any external dependencies, useful for development
        and testing package structure.

        Run:

        .. code-block:: bash

            poetry install --only-root

        Ref:

        - poetry install: https://python-poetry.org/docs/cli/#install
        """
        return self._run_poetry_command(
            args=["install", "--only-root"],
            real_run=real_run,
            quiet=quiet,
        )

    def poetry_install_only_root(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(disable=not verbose):
            return self._poetry_install_only_root(
                real_run=real_run,
                quiet=not verbose,
            )

    poetry_install_only_root.__doc__ = _poetry_install_only_root.__doc__

    @logger.emoji_block(
        msg="Install main dependencies and Package itself",
        emoji=Emoji.install,
    )
    def _poetry_install(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Install main dependencies and the package in editable mode.

        Installs core project dependencies without development or optional groups.

        Run:

        .. code-block:: bash

            poetry install

        Ref:

        - poetry install: https://python-poetry.org/docs/cli/#install
        """
        return self._run_poetry_command(
            args=["install"],
            real_run=real_run,
            quiet=quiet,
        )

    def poetry_install(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(disable=not verbose):
            return self._poetry_install(
                real_run=real_run,
                quiet=not verbose,
            )

    poetry_install.__doc__ = _poetry_install.__doc__

    @logger.emoji_block(
        msg="Install dev dependencies",
        emoji=Emoji.install,
    )
    def _poetry_install_dev(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Install development dependencies. Adds dependencies from the
        ``[tool.poetry.group.dev.dependencies]`` group to the project environment.

        Run:

        .. code-block:: bash

            poetry install --extras dev

        Ref:

        - poetry install: https://python-poetry.org/docs/cli/#install
        """
        return self._run_poetry_command(
            args=["install", "--extras", "dev"],
            real_run=real_run,
            quiet=quiet,
        )

    def poetry_install_dev(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(disable=not verbose):
            return self._poetry_install_dev(
                real_run=real_run,
                quiet=not verbose,
            )

    poetry_install_dev.__doc__ = _poetry_install_dev.__doc__

    @logger.emoji_block(
        msg="Install test dependencies",
        emoji=Emoji.install,
    )
    def _poetry_install_test(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Install test dependencies. Adds dependencies from the
        ``[tool.poetry.group.test.dependencies]`` group to the project environment.

        Run:

        .. code-block:: bash

            poetry install --extras test

        Ref:

        - poetry install: https://python-poetry.org/docs/cli/#install
        """
        return self._run_poetry_command(
            args=["install", "--extras", "test"],
            real_run=real_run,
            quiet=quiet,
        )

    def poetry_install_test(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(disable=not verbose):
            return self._poetry_install_test(
                real_run=real_run,
                quiet=not verbose,
            )

    poetry_install_test.__doc__ = _poetry_install_test.__doc__

    @logger.emoji_block(
        msg="Install doc dependencies",
        emoji=Emoji.install,
    )
    def _poetry_install_doc(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Install documentation build dependencies. Adds dependencies from the
        ``[tool.poetry.group.doc.dependencies]`` group to the project environment.

        Run:

        .. code-block:: bash

            poetry install --extras doc

        Ref:

        - poetry install: https://python-poetry.org/docs/cli/#install
        """
        return self._run_poetry_command(
            args=["install", "--extras", "doc"],
            real_run=real_run,
            quiet=quiet,
        )

    def poetry_install_doc(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(disable=not verbose):
            return self._poetry_install_doc(
                real_run=real_run,
                quiet=not verbose,
            )

    poetry_install_doc.__doc__ = _poetry_install_doc.__doc__

    @logger.emoji_block(
        msg="Install automation dependencies",
        emoji=Emoji.install,
    )
    def _poetry_install_auto(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Install automation dependencies. Adds dependencies from the
        ``[tool.poetry.group.auto.dependencies]`` group to the project environment.

        Run:

        .. code-block:: bash

            poetry install --extras auto

        Ref:

        - poetry install: https://python-poetry.org/docs/cli/#install
        """
        return self._run_poetry_command(
            args=["install", "--extras", "auto"],
            real_run=real_run,
            quiet=quiet,
        )

    def poetry_install_auto(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(disable=not verbose):
            return self._poetry_install_auto(
                real_run=real_run,
                quiet=not verbose,
            )

    poetry_install_auto.__doc__ = _poetry_install_auto.__doc__

    @logger.emoji_block(
        msg="Install all dependencies for dev, test, doc",
        emoji=Emoji.install,
    )
    def _poetry_install_all(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Install all dependency groups.

        Adds dependencies from all configured groups to the project environment.

        Run:

        .. code-block:: bash

            poetry install --all-extras

        Ref:

        - poetry install: https://python-poetry.org/docs/cli/#install
        """
        return self._run_poetry_command(
            args=["install", "--all-extras"],
            real_run=real_run,
            quiet=quiet,
        )

    def poetry_install_all(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(disable=not verbose):
            return self._poetry_install_all(
                real_run=real_run,
                quiet=not verbose,
            )

    poetry_install_all.__doc__ = _poetry_install_all.__doc__

    def _do_we_need_poetry_export(
        self: "PyWf",
        current_poetry_lock_hash: str,
    ) -> bool:  # pragma: no cover
        """
        The ``poetry export`` command is resource-intensive, so we use a
        caching mechanism to avoid unnecessary executions.

        Each time we run :meth:`PyWfDeps._poetry_export`, it calculates the
        SHA-256 hash of the current ``poetry.lock`` file and writes it to
        a cache file named ``poetry-lock-hash.json``, located at the root of the repository.

        Before exporting, the function compares the current hash of ``poetry.lock``
        with the value stored in the cache file. If the hashes differ, it indicates
        that ``poetry.lock`` has changed, and we should run :meth:`PyWfDeps._poetry_export` again.

        An example of the ``poetry-lock-hash.json`` file content::

            {
                "hash": "sha256-hash-of-the-poetry.lock-file",
                "description": "DON'T edit this file manually!"
            }

        Ref:

        - poetry export: https://python-poetry.org/docs/cli/#export

        :param current_poetry_lock_hash: the sha256 hash of the current ``poetry.lock`` file
        """
        if self.path_poetry_lock_hash_json.exists():
            # read the previous poetry lock hash from cache file
            cached_poetry_lock_hash = json.loads(
                self.path_poetry_lock_hash_json.read_text()
            )["hash"]
            return current_poetry_lock_hash != cached_poetry_lock_hash
        else:
            # do poetry export if the cache file not found
            return True

    def _poetry_export_main(
        self: "PyWf",
        with_hash: bool = False,
        real_run: bool = True,
    ):  # pragma: no cover
        """
        Export main project dependencies to ``requirements.txt``.

        :param with_hash: whether to include the hash of the dependencies in the
            requirements.txt file.
        """
        self.path_requirements.unlink(missing_ok=True)
        args = [
            f"{self.path_bin_poetry}",
            "export",
            "--format",
            "requirements.txt",
            "--output",
            f"{self.path_requirements}",
        ]
        if with_hash is False:
            args.append("--without-hashes")
        print_command(args)
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)

    def _poetry_export_group(
        self: "PyWf",
        group: str,
        path: Path,
        with_hash: bool = False,
        real_run: bool = True,
    ):  # pragma: no cover
        """
        Export specific dependency group to ``requirements-{group}.txt``.

        Ref: https://github.com/python-poetry/poetry-plugin-export

        :param group: dependency group name, for example dev dependencies are defined
            in the ``[tool.poetry.group.dev]`` and ``[tool.poetry.group.dev.dependencies]``
            sections of he ``pyproject.toml`` file.
        :param path: the path to the exported ``requirements.txt`` file.
        :param with_hash: whether to include the hash of the dependencies in the
            ``requirements.txt`` file.
        """
        if real_run:
            path.unlink(missing_ok=True)

        args = [
            f"{self.path_bin_poetry}",
            "export",
            "--format",
            "requirements.txt",
            "--output",
            f"{path}",
            "--extras",
        ]
        args.append(group)
        if with_hash is False:
            args.append("--without-hashes")
        print_command(args)
        if real_run:
            with temp_cwd(self.dir_project_root):
                subprocess.run(args, check=True)

    def _poetry_export_logic(
        self: "PyWf",
        current_poetry_lock_hash: str,
        with_hash: bool = False,
        real_run: bool = True,
    ):  # pragma: no cover
        """
        Run ``poetry export --format requirements.txt ...`` command and write
        the sha256 hash of the current ``poetry.lock`` file to the cache file.

        Run:

        .. code-block:: bash

            poetry export --format requirements.txt --output requirements.txt --without-hashes
            poetry export --format requirements.txt --output requirements-dev.txt --extras dev --without-hashes
            poetry export --format requirements.txt --output requirements-test.txt --extras test --without-hashes
            poetry export --format requirements.txt --output requirements-doc.txt --extras doc --without-hashes
            poetry export --format requirements.txt --output requirements-automation.txt --extras auto --without-hashes

        :param current_poetry_lock_hash: the sha256 hash of the current ``poetry.lock`` file
        :param with_hash: whether to include the hash of the dependencies in the
            requirements.txt file.
        """
        # export the main dependencies
        self._poetry_export_main(with_hash=with_hash, real_run=real_run)

        # export dev, test, doc, auto dependencies
        for group, path in [
            ("dev", self.path_requirements_dev),
            ("test", self.path_requirements_test),
            ("doc", self.path_requirements_doc),
            ("auto", self.path_requirements_automation),
        ]:
            self._poetry_export_group(
                group=group,
                path=path,
                with_hash=with_hash,
                real_run=real_run,
            )

        # write the ``poetry.lock`` hash to the cache file
        if real_run:
            self.path_poetry_lock_hash_json.write_text(
                json.dumps(
                    {
                        "hash": current_poetry_lock_hash,
                        "description": (
                            "DON'T edit this file manually! This file is the cache of "
                            "the poetry.lock file hash. It is used to avoid unnecessary "
                            "expansive 'poetry export ...' command."
                        ),
                    },
                    indent=4,
                )
            )

    @logger.emoji_block(
        msg="Export all dependencies to requirements-***.txt",
        emoji=Emoji.install,
    )
    def _poetry_export(
        self: "PyWf",
        with_hash: bool = False,
        real_run: bool = True,
        quiet: bool = False,
    ) -> bool:
        """
        :return: ``True`` if ``poetry export`` is executed, ``False`` if not.
        """
        poetry_lock_hash = sha256_of_bytes(self.path_poetry_lock.read_bytes())
        if self._do_we_need_poetry_export(poetry_lock_hash):
            self._poetry_export_logic(
                current_poetry_lock_hash=poetry_lock_hash,
                with_hash=with_hash,
                real_run=real_run,
            )
            return True
        else:
            logger.info("already did, do nothing")
            return False

    def poetry_export(
        self: "PyWf",
        with_hash: bool = False,
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(disable=not verbose):
            flag = self._poetry_export(
                with_hash=with_hash,
                real_run=real_run,
                quiet=not verbose,
            )
            return flag

    poetry_export.__doc__ = _poetry_export_logic.__doc__



# -*- coding: utf-8 -*-

"""
Publish to Python repository related automation.
"""

import typing as T
import subprocess
import dataclasses
from textwrap import dedent

try:
    from github import (
        Github,
        GithubException,
        Repository,
        GitTag,
        GitRelease,
    )
except ImportError:  # pragma: no cover
    pass

from .vendor.emoji import Emoji
from .vendor.better_pathlib import temp_cwd

from .logger import logger
from .helpers import bump_version, print_command

if T.TYPE_CHECKING:  # pragma: no cover
    from .define import PyWf


@dataclasses.dataclass
class PyWfPublish:  # pragma: no cover
    """
    Namespace class for publishing to Python repository related automation.
    """
    def bump_version(
        self: "PyWf",
        major: bool = False,
        minor: bool = False,
        patch: bool = False,
        minor_start_from: int = 0,
        micro_start_from: int = 0,
        real_run: bool = True,
    ):
        """
        Bump a semantic version. The current version has to be in x.y.z format,
        where x, y, z are integers.

        :param major: bump major version.
        :param minor: bump minor version.
        :param patch: bump patch version.
        :param minor_start_from: if bumping major version, minor start from this number.
        :param micro_start_from: if bumping minor version, micro start from this number.
        """
        new_version = bump_version(
            current_version=self.package_version,
            major=major,
            minor=minor,
            patch=patch,
            minor_start_from=minor_start_from,
            micro_start_from=micro_start_from,
        )

        # update _version.py file
        version_py_content = dedent(
            """
        __version__ = "{}"
    
        # keep this ``if __name__ == "__main__"``, don't delete!
        # this is used by automation script to detect the project version
        if __name__ == "__main__":  # pragma: no cover
            print(__version__)
        """
        ).strip()
        version_py_content = version_py_content.format(new_version)
        if real_run:
            self.path_version_py.write_text(version_py_content)

        # update pyproject.toml file
        if self.path_pyproject_toml.exists():
            if major:
                action = "major"
            elif minor:
                action = "minor"
            elif patch:
                action = "patch"
            else:  # pragma: no cover
                raise NotImplementedError
            with temp_cwd(self.dir_project_root):
                args = [
                    f"{self.path_bin_poetry}",
                    "version",
                    action,
                ]
                print_command(args)
                if real_run:
                    subprocess.run(args, check=True)

    @logger.emoji_block(
        msg="Publish to GitHub Release",
        emoji=Emoji.package,
    )
    def _publish_to_github_release(
        self: "PyWf",
        real_run: bool = True,
    ) -> T.Optional["GitRelease"]:
        """
        Create a GitHub Release using the current version based on main branch.

        :returns: a boolean flag to indicate whether the operation is performed.
        """
        logger.info(f"preview release at {self.github_versioned_release_url}")

        release_name = self.package_version
        gh = Github(self.github_token)
        repo = gh.get_repo(self.github_repo_fullname)

        # Check if release exists
        try:
            repo.get_release(release_name)
            logger.info(f"Release {release_name!r} already exists.")
            return None
        except GithubException as e:
            if e.status == 404:
                pass
            else:
                raise e
        except Exception as e:  # pragma: no cover
            raise e

        # Create Tag if not exists
        try:
            repo.get_git_ref(f"tags/{release_name}")
            logger.info(f"Tag {release_name!r} already exists.")
        except GithubException as e:
            if e.status == 404:
                if real_run:
                    default_branch = repo.default_branch
                    commit = repo.get_branch(default_branch).commit
                    commit_sha = commit.sha
                    tag = repo.create_git_tag(
                        tag=release_name,
                        message=f"Release {release_name}",
                        object=commit_sha,
                        type="commit",
                    )
                    repo.create_git_ref(
                        ref=f"refs/tags/{release_name}",
                        sha=tag.sha,
                    )
            else:  # pragma: no cover
                raise e
        except Exception as e:  # pragma: no cover
            raise e

        # Create Release
        if real_run:
            release = repo.create_git_release(
                tag=release_name,
                name=release_name,
                message=f"Release {release_name}",
            )
            return release
        else:
            return None

    def publish_to_github_release(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):
        with logger.disabled(not verbose):
            return self._publish_to_github_release(
                real_run=real_run,
            )

    publish_to_github_release.__doc__ = _publish_to_github_release.__doc__



# -*- coding: utf-8 -*-

"""
Enumeration of important paths on local file system.
"""

import typing as T
import sys
import subprocess
import dataclasses
from pathlib import Path
from functools import cached_property

if T.TYPE_CHECKING:  # pragma: no cover
    from .define import PyWf


@dataclasses.dataclass
class PyWfPaths:
    """
    Namespace class for accessing important paths.
    """

    @cached_property
    def dir_home(self: "PyWf") -> Path:
        """
        The user home directory.

        Example: ``${HOME}``
        """
        return Path.home()

    # --------------------------------------------------------------------------
    # Virtualenv
    # --------------------------------------------------------------------------
    _VENV_RELATED = None

    @property
    def dir_venv(self: "PyWf") -> Path:
        """
        The virtualenv directory.

        Example: ``${dir_project_root}/.venv``
        """
        return self.dir_project_root.joinpath(".venv")

    @property
    def dir_venv_bin(self: "PyWf") -> Path:
        """
        The bin folder in virtualenv.

        Example: ``${dir_project_root}/.venv/bin``
        """
        return self.dir_venv.joinpath("bin")

    def get_path_venv_bin_cli(self, cmd: str) -> Path:
        """
        Get the path of a command in virtualenv bin folder.

        Example: ``${dir_project_root}/.venv/bin/${cmd}``
        """
        return self.dir_venv_bin.joinpath(cmd)

    @property
    def path_venv_bin_python(self: "PyWf") -> Path:
        """
        The python executable in virtualenv.

        Example: ``${dir_project_root}/.venv/bin/python``
        """
        return self.get_path_venv_bin_cli("python")

    @property
    def path_venv_bin_pip(self: "PyWf") -> Path:
        """
        The pip command in virtualenv.

        Example: ``${dir_project_root}/.venv/bin/pip``
        """
        return self.get_path_venv_bin_cli("pip")

    @property
    def path_venv_bin_pytest(self: "PyWf") -> Path:
        """
        The pytest command in virtualenv.

        Example: ``${dir_project_root}/.venv/bin/pytest``
        """
        return self.get_path_venv_bin_cli("pytest")

    @property
    def path_venv_bin_sphinx_build(self: "PyWf") -> Path:
        """
        The sphinx-build executable in virtualenv.

        Example: ``${dir_project_root}/.venv/bin/sphinx-build``
        """
        return self.get_path_venv_bin_cli("sphinx-build")

    @property
    def path_sys_executable(self: "PyWf") -> Path:
        """
        The current Python interpreter path.
        """
        return Path(sys.executable)

    def get_path_dynamic_bin_cli(self, cmd: str) -> Path:
        """
        Search multiple locations to get the absolute path of the CLI command.
        It searches the following locations in order:

        1. the bin folder in virtualenv.
        2. the global Python's bin folder.
        3. Then use the raw command name (string) as the path.

        Example: ``${dir_project_root}/.venv/bin/${cmd}`` or ``${global_python_bin}/${cmd}``
        """
        p = self.dir_venv_bin.joinpath(cmd)
        if p.exists():
            return p
        p = self.path_sys_executable.parent.joinpath(cmd)
        if p.exists():
            return p
        return Path(cmd)

    @property
    def path_bin_virtualenv(self: "PyWf") -> Path:
        """
        The virtualenv CLI command path.

        Example: ``${dir_project_root}/.venv/bin/virtualenv``
        """
        return self.get_path_dynamic_bin_cli("virtualenv")

    @property
    def path_bin_poetry(self: "PyWf") -> Path:
        """
        The poetry CLI command path.

        Example: ``${dir_project_root}/.venv/bin/poetry``
        """
        return self.get_path_dynamic_bin_cli("poetry")

    @property
    def path_bin_twine(self: "PyWf") -> Path:
        """
        The twine CLI command path.

        Example: ``${dir_project_root}/.venv/bin/twine``
        """
        return self.get_path_dynamic_bin_cli("twine")

    # --------------------------------------------------------------------------
    # Source code
    # --------------------------------------------------------------------------
    @property
    def dir_python_lib(self: "PyWf") -> Path:
        """
        The current Python library directory.

        Example: ``${dir_project_root}/${package_name}``
        """
        return self.dir_project_root.joinpath(self.package_name)

    @property
    def path_version_py(self: "PyWf") -> Path:
        """
        Path to the ``_version.py`` file where the package version is defined.

        Example: ``${dir_project_root}/${package_name}/_version.py``
        """
        return self.dir_python_lib.joinpath("_version.py")

    # --------------------------------------------------------------------------
    # Pytest
    # --------------------------------------------------------------------------
    _PYTEST_RELATED = None

    @property
    def dir_tests(self: "PyWf") -> Path:
        """
        Unit test folder.

        Example: ``${dir_project_root}/tests``
        """
        return self.dir_project_root.joinpath("tests")

    @property
    def dir_tests_int(self: "PyWf") -> Path:
        """
        Integration test folder.

        Example: ``${dir_project_root}/tests_int``
        """
        return self.dir_project_root.joinpath("tests_int")

    @property
    def dir_tests_load(self: "PyWf") -> Path:
        """
        Load test folder.

        Example: ``${dir_project_root}/tests_load``
        """
        return self.dir_project_root.joinpath("tests_load")

    @property
    def dir_htmlcov(self: "PyWf") -> Path:
        """
        The code coverage test results HTML output folder.

        Example: ``${dir_project_root}/htmlcov``
        """
        return self.dir_project_root.joinpath("htmlcov")

    @property
    def path_htmlcov_index_html(self: "PyWf") -> Path:
        """
        The code coverage test results HTML file.

        Example: ``${dir_project_root}/htmlcov/index.html``
        """
        return self.dir_htmlcov.joinpath("index.html")

    # --------------------------------------------------------------------------
    # Sphinx doc
    # --------------------------------------------------------------------------
    _SPHINX_DOC_RELATED = None

    @property
    def dir_sphinx_doc(self: "PyWf") -> Path:
        """
        Sphinx docs folder.

        Example: ``${dir_project_root}/docs``
        """
        return self.dir_project_root.joinpath("docs")

    @property
    def dir_sphinx_doc_source(self: "PyWf") -> Path:
        """
        Sphinx docs source code folder.

        Example: ``${dir_project_root}/docs/source``
        """
        return self.dir_sphinx_doc.joinpath("source")

    @property
    def dir_sphinx_doc_source_conf_py(self: "PyWf") -> Path:
        """
        Sphinx docs ``conf.py`` file path.

        Example: ``${dir_project_root}/docs/source/conf.py``
        """
        return self.dir_sphinx_doc_source.joinpath("conf.py")

    @property
    def dir_sphinx_doc_source_python_lib(self: "PyWf") -> Path:
        """
        The generated Python library API reference Sphinx docs folder.

        Example: ``${dir_project_root}/docs/source/${package_name}``
        """
        return self.dir_sphinx_doc_source.joinpath(self.package_name)

    @property
    def dir_sphinx_doc_build(self: "PyWf") -> Path:
        """
        The temp Sphinx doc build folder.

        Example: ``${dir_project_root}/docs/build
        """
        return self.dir_sphinx_doc.joinpath("build")

    @property
    def dir_sphinx_doc_build_html(self: "PyWf") -> Path:
        """
        The built Sphinx doc build HTML folder.

        Example: ``${dir_project_root}/docs/build/html
        """
        return self.dir_sphinx_doc_build.joinpath("html")

    @property
    def path_sphinx_doc_build_index_html(self: "PyWf") -> Path:  # pragma: no cover
        """
        The built Sphinx doc site entry HTML file path.

        Example: ``${dir_project_root}/docs/build/html/index.html or README.html
        """
        if self.dir_sphinx_doc_source.joinpath("index.rst").exists():
            return self.dir_sphinx_doc_build_html.joinpath("index.html")

        if self.dir_sphinx_doc_source.joinpath("README.rst").exists():
            return self.dir_sphinx_doc_build_html.joinpath("README.html")

        raise FileNotFoundError(
            str(self.dir_sphinx_doc_build_html.joinpath("index.html"))
        )

    # --------------------------------------------------------------------------
    # Poetry
    # --------------------------------------------------------------------------
    _POETRY_RELATED = None

    @property
    def path_requirements(self: "PyWf") -> Path:
        """
        The requirements.txt file path.

        Example: ``${dir_project_root}/requirements.txt``
        """
        return self.dir_project_root.joinpath("requirements.txt")

    @property
    def path_requirements_dev(self: "PyWf") -> Path:
        """
        The requirements-dev.txt file path.

        Example: ``${dir_project_root}/requirements-dev.txt``
        """
        return self.dir_project_root.joinpath("requirements-dev.txt")

    @property
    def path_requirements_test(self: "PyWf") -> Path:
        """
        The requirements-test.txt file path.

        Example: ``${dir_project_root}/requirements-test.txt``
        """
        return self.dir_project_root.joinpath("requirements-test.txt")

    @property
    def path_requirements_doc(self: "PyWf") -> Path:
        """
        The requirements-doc.txt file path.

        Example: ``${dir_project_root}/requirements-doc.txt``
        """
        return self.dir_project_root.joinpath("requirements-doc.txt")

    @property
    def path_requirements_automation(self: "PyWf") -> Path:
        """
        The requirements-automation.txt file path.

        Example: ``${dir_project_root}/requirements-automation.txt``
        """
        return self.dir_project_root.joinpath("requirements-automation.txt")

    @property
    def path_poetry_lock(self: "PyWf") -> Path:
        """
        The poetry.lock file path.

        Example: ``${dir_project_root}/poetry.lock``
        """
        return self.dir_project_root.joinpath("poetry.lock")

    @property
    def path_poetry_lock_hash_json(self: "PyWf") -> Path:
        """
        The poetry-lock-hash.json file path. It is the cache of the poetry.lock file hash.

        Example: ``${dir_project_root}/poetry-lock-hash.json``
        """
        return self.dir_project_root.joinpath("poetry-lock-hash.json")

    # ------------------------------------------------------------------------------
    # Build Related
    # ------------------------------------------------------------------------------
    _BUILD_RELATED = None

    @property
    def path_pyproject_toml(self: "PyWf") -> Path:
        """
        The pyproject.toml file path.

        Example: ``${dir_project_root}/pyproject.toml``
        """
        return self.dir_project_root.joinpath("pyproject.toml")

    @property
    def dir_build(self: "PyWf") -> Path:
        """
        The build folder for Python or artifacts build.

        Example: ``${dir_project_root}/build``
        """
        return self.dir_project_root.joinpath("build")

    @property
    def dir_dist(self: "PyWf") -> Path:
        """
        The dist folder for Python package distribution (.whl file).

        Example: ``${dir_project_root}/dist``
        """
        return self.dir_project_root.joinpath("dist")

    # --------------------------------------------------------------------------
    # Token Files
    # --------------------------------------------------------------------------
    @property
    def path_github_token_file(self: "PyWf") -> Path:
        """
        Create GitHub token in https://github.com/settings/tokens and put the token at
        ``${HOME}/.github/${github_account}/pac/${github_token_name}.txt``
        """
        return self.dir_home.joinpath(
            ".github",
            self.github_account,
            "pac",
            f"{self.github_token_name}.txt",
        )

    @property
    def path_codecov_token_file(self: "PyWf") -> Path:
        """
        Create Codecov token in https://app.codecov.io/account/gh/${codecov_account}/access and put the token at
        ``${HOME}/.codecov/github/${codecov_account}/${codecov_token_name}.txt``

        .. note::

            If you use Github to login, then codecov_account is your github_account.
        """
        return self.dir_home.joinpath(
            ".codecov",
            "github",
            self.codecov_account,
            f"{self.codecov_token_name}.txt",
        )

    @property
    def path_cloudflare_token_file(self: "PyWf") -> Path:
        """
        Create CloudFlare User API token in https://dash.cloudflare.com/profile/api-tokens
        Give it the following permission:

        - All accounts - Cloudflare Pages:Read, Cloudflare Pages:Edit
        - All users - Memberships:Read, User Details:Read

        And put the token at ``${HOME}/.cloudflare/${cloudflare_account_alias}/{cloudflare_pages_token_name}.txt``
        """
        return self.dir_home.joinpath(
            ".cloudflare",
            self.cloudflare_account_alias,
            f"{self.cloudflare_pages_token_name}.txt",
        )

    # --------------------------------------------------------------------------
    # AWS Related
    # --------------------------------------------------------------------------
    _AWS_RELATED = None

    @property
    def path_bin_aws(self: "PyWf") -> Path:
        """
        The AWS CLI executable path.

        Example: ``${dir_project_root}/.venv/bin/aws``
        """
        return self.get_path_dynamic_bin_cli("aws")

    # --------------------------------------------------------------------------
    # Cloudflare
    # --------------------------------------------------------------------------
    _CLOUDFLARE_RELATED = None

    @property
    def dir_node_modules(self: "PyWf") -> Path:
        """
        """
        return self.dir_home.joinpath(".npm-tools", "node_modules")

    @property
    def path_bin_wrangler(self: "PyWf") -> Path:
        """
        The `Cloudflare wrangler <https://developers.cloudflare.com/workers/wrangler/install-and-update/>`_
        CLI command path.
        
        Suggestion:
        
        .. code-block:: bash
        
            mkdir ${HOME}/.npm-tools
            cd ${HOME}/.npm-tools
            npm install wrangler --save-dev
        """
        return self.dir_node_modules.joinpath(".bin", "wrangler")



# -*- coding: utf-8 -*-

"""
Build Python source distribution.

Usage example:

.. code-block:: python

    from better_pathlib import temp_cwd
"""

import typing as T
import os
import contextlib
from pathlib import Path

__version__ = "0.1.1"


@contextlib.contextmanager
def temp_cwd(path: T.Union[str, Path]):
    """
    Temporarily set the current working directory (CWD) and automatically
    switch back when it's done.

    Example:

    .. code-block:: python

        with temp_cwd(Path("/path/to/target/working/directory")):
            # do something
    """
    path = Path(path).absolute()
    if not path.is_dir():
        raise NotADirectoryError(f"{path} is not a dir!")
    cwd = os.getcwd()
    os.chdir(str(path))
    try:
        yield path
    finally:
        os.chdir(cwd)


def get_dir_here(file_var: str) -> Path:
    """
    Get the directory of the current file.

    Example:

    .. code-block:: python

        get_dir_here(__file__)
    """
    return Path(file_var).absolute().parent



# -*- coding: utf-8 -*-

"""
Virtualenv management related automation.
"""

import typing as T
import shutil
import subprocess
import dataclasses

from .vendor.emoji import Emoji
from .vendor.better_pathlib import temp_cwd

from .logger import logger
from .helpers import print_command

if T.TYPE_CHECKING:  # pragma: no cover
    from .define import PyWf


@dataclasses.dataclass
class PyWfVenv:
    """
    Namespace class for Virtualenv management related automation.
    """

    @logger.emoji_block(
        msg="Create Virtual Environment",
        emoji=Emoji.python,
    )
    def _create_virtualenv(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ) -> bool:
        """
        Run:

        .. code-block:: bash

            $ poetry env use python${X}.${Y}

        :return: a boolean flat to indicate whether a creation is performed.
        """
        if self.dir_venv.exists():
            logger.info(f"{self.dir_venv} already exists, do nothing.")
            return False
        else:
            # Ref: https://python-poetry.org/docs/managing-environments/
            # note that we defined to use in-project = true in poetry.toml file
            args = [
                f"{self.path_bin_poetry}",
                "config",
                "virtualenvs.in-project",
                "true",
            ]
            if quiet:
                args.append("--quiet")
            print_command(args)

            if real_run is True:
                subprocess.run(args, check=True)
            args = [
                f"{self.path_bin_poetry}",
                "env",
                "use",
                f"python{self.py_ver_major}.{self.py_ver_minor}",
            ]
            if quiet:
                args.append("--quiet")
            print_command(args)
            if real_run is True:
                with temp_cwd(self.dir_project_root):
                    subprocess.run(args, check=True)
            logger.info("done")
            return True

    def create_virtualenv(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ) -> bool:  # pragma: no cover
        with logger.disabled(not verbose):
            return self._create_virtualenv(
                real_run=real_run,
                quiet=not verbose,
            )

    create_virtualenv.__doc__ = _create_virtualenv.__doc__

    @logger.emoji_block(
        msg="Remove Virtual Environment",
        emoji=Emoji.python,
    )
    def _remove_virtualenv(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ) -> bool:
        """
        Run:

        .. code-block:: bash

            $ rm -r /path/to/.venv

        :return: a boolean flag to indicate whether a deletion is performed.
        """
        if self.dir_venv.exists():
            args = [
                "rm",
                "-r",
                f"{self.dir_venv}",
            ]
            print_command(args)
            if real_run:
                shutil.rmtree(f"{self.dir_venv}", ignore_errors=True)
            logger.info(f"done! {self.dir_venv} is removed.")
            return True
        else:
            logger.info(f"{self.dir_venv} doesn't exists, do nothing.")
            return False

    def remove_virtualenv(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):
        with logger.disabled(not verbose):
            return self._remove_virtualenv(
                real_run=real_run,
                quiet=not verbose,
            )

    remove_virtualenv.__doc__ = _remove_virtualenv.__doc__


# -*- coding: utf-8 -*-

"""
Source Code Build Automation for Python Projects.
"""

import typing as T
import shutil
import dataclasses

from .vendor.emoji import Emoji
from .vendor.build_dist import (
    build_dist_with_python_build,
    build_dist_with_poetry_build,
)

from .logger import logger

if T.TYPE_CHECKING:  # pragma: no cover
    from .define import PyWf


@dataclasses.dataclass
class PyWfBuild:
    """
    Namespace class for build related automation.
    """

    @logger.emoji_block(
        msg="Build python distribution using pypa-build",
        emoji=Emoji.build,
    )
    def _python_build(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Build python source distribution using
        `pypa-build <https://pypa-build.readthedocs.io/en/latest/>`_.
        
        Run:
        
        .. code-block:: bash
            
            python -m build --sdist --wheel
        """
        if self.dir_dist.exists():
            if real_run:
                shutil.rmtree(self.dir_dist, ignore_errors=True)
        build_dist_with_python_build(
            dir_project_root=self.dir_project_root,
            path_bin_python=self.path_venv_bin_python,
            log_func=logger.info,
            real_run=real_run,
            verbose=not quiet,
        )

    def python_build(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(not verbose):
            return self._python_build(
                real_run=real_run,
                quiet=not verbose,
            )

    python_build.__doc__ = _python_build.__doc__

    @logger.emoji_block(
        msg="Build python distribution using poetry",
        emoji=Emoji.build,
    )
    def _poetry_build(
        self: "PyWf",
        real_run: bool = True,
        quiet: bool = False,
    ):
        """
        Build python source distribution using

        `poetry build <https://python-poetry.org/docs/cli/#build>`_.

        Run:

        .. code-block:: bash

            poetry build
        """
        if self.dir_dist.exists():
            if real_run:
                shutil.rmtree(self.dir_dist, ignore_errors=True)
        build_dist_with_poetry_build(
            dir_project_root=self.dir_project_root,
            path_bin_poetry=self.path_bin_poetry,
            log_func=logger.info,
            real_run=real_run,
            verbose=not quiet,
        )

    def poetry_build(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):  # pragma: no cover
        with logger.disabled(not verbose):
            return self._poetry_build(
                real_run=real_run,
                quiet=not verbose,
            )

    poetry_build.__doc__ = _poetry_build.__doc__



# -*- coding: utf-8 -*-

"""
Setup SaaS services for your Open Source Python project.
"""

import typing as T
import dataclasses
from functools import cached_property


try:
    import requests
    from github import Github
except ImportError:  # pragma: no cover
    pass

from .vendor.emoji import Emoji

from .logger import logger
from .helpers import raise_http_response_error

if T.TYPE_CHECKING:  # pragma: no cover
    from .define import PyWf


@dataclasses.dataclass
class PyWfSaas:  # pragma: no cover
    """
    Namespace class for SaaS service setup automation.
    """

    @cached_property
    def github_token(self: "PyWf") -> str:
        if self.path_github_token_file.exists():
            return self.path_github_token_file.read_text(encoding="utf-8").strip()
        else:  # pragma: no cover
            message = (
                f"{Emoji.error} Cannot find GitHub token file at "
                f"{self.path_github_token_file}!\n"
                f"{self.__class__.path_github_token_file.__doc__}"
            )
            raise FileNotFoundError(message)

    @cached_property
    def codecov_token(self: "PyWf") -> str:
        if self.path_codecov_token_file.exists():
            return self.path_codecov_token_file.read_text(encoding="utf-8").strip()
        else:  # pragma: no cover
            message = (
                f"{Emoji.error} Cannot find Codecov token file at "
                f"{self.path_codecov_token_file}!\n"
                f"{self.__class__.path_codecov_token_file.__doc__}"
            )
            raise FileNotFoundError(message)

    @cached_property
    def readthedocs_token(self: "PyWf") -> str:
        if self.path_readthedocs_token_file.exists():
            return self.path_readthedocs_token_file.read_text(encoding="utf-8").strip()
        else:
            message = (
                f"{Emoji.error} Cannot find Readthedocs token file at "
                f"{self.path_readthedocs_token_file}!\n"
                f"{self.__class__.path_readthedocs_token_file.__doc__}"
            )
            raise FileNotFoundError(message)

    def get_codecov_io_upload_token(
        self: "PyWf",
        real_run: bool = True,
    ) -> T.Optional[str]:
        """
        Get the upload token for codecov io for your GitHub repo.

        Ref:

        - https://docs.codecov.com/reference/repos_retrieve
        - https://docs.codecov.com/reference/repos_config_retrieve

        :returns: the upload token for codecov.io.
        """
        logger.info("Getting codecov.io upload token...")
        url = f"https://app.codecov.io/gh/{self.github_account}/{self.git_repo_name}/settings"
        with logger.indent():
            logger.info(f"preview at {url}")
        headers = {
            "accept": "application/json",
            "authorization": f"Bearer {self.codecov_token}",
        }
        endpoint = "https://api.codecov.io/api/v2"
        url = f"{endpoint}/github/{self.github_account}/repos/{self.git_repo_name}/"
        if real_run:
            response = requests.get(url, headers=headers)
            response.raise_for_status()
            is_private = response.json()["private"]
            if is_private is True:
                raise ValueError("You cannot use codecov.io for private repositories.")

        url = f"{endpoint}/github/{self.github_account}/repos/{self.git_repo_name}/config/"
        if real_run:
            response = requests.get(url, headers=headers)
            response.raise_for_status()
            upload_token = response.json()["upload_token"]
            return upload_token
        else:
            return None

    @logger.emoji_block(
        msg="Setup codecov.io Upload Token on GitHub",
        emoji=Emoji.test,
    )
    def _setup_codecov_io_upload_token_on_github(
        self: "PyWf",
        real_run: bool = True,
    ):
        """
        Apply the codecov upload token to GitHub Action secrets in your GitHub repository.

        Ref:

        - https://docs.codecov.com/reference/repos_retrieve
        - https://docs.codecov.com/reference/repos_config_retrieve
        - https://pygithub.readthedocs.io/en/latest/examples/Repository.html

        :returns: a boolean flag to indicate whether the operation is performed.
        """
        codecov_io_upload_token = self.get_codecov_io_upload_token(real_run=real_run)

        logger.info("Setting up codecov.io upload token on GitHub...")
        with logger.indent():
            logger.info(f"preview at {self.github_actions_secrets_settings_url}")
        gh = Github(self.github_token)
        repo = gh.get_repo(self.github_repo_fullname)
        if real_run:
            repo.create_secret(
                secret_name="CODECOV_TOKEN",
                unencrypted_value=codecov_io_upload_token,
                secret_type="actions",
            )
        return real_run

    def setup_codecov_io_upload_token_on_github(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):
        with logger.disabled(not verbose):
            return self._setup_codecov_io_upload_token_on_github(
                real_run=real_run,
            )

    setup_codecov_io_upload_token_on_github.__doc__ = (
        _setup_codecov_io_upload_token_on_github.__doc__
    )

    @logger.emoji_block(
        msg="Setup readthedocs.org Project",
        emoji=Emoji.doc,
    )
    def _setup_readthedocs_project(
        self: "PyWf",
        real_run: bool = True,
    ) -> bool:
        """
        Create a project on readthedocs.org.

        Ref:

        - https://docs.readthedocs.io/en/stable/api/v3.html#get--api-v3-projects-(string-project_slug)-
        - https://docs.readthedocs.io/en/stable/api/v3.html#post--api-v3-projects-

        :returns: a boolean flag to indicate whether the operation is performed.
        """
        logger.info("Setting up readthedocs project...")
        readthedocs_project_name_slug = self.readthedocs_project_name.replace("_", "-")
        url = f"https://app.readthedocs.org/dashboard/{readthedocs_project_name_slug}/edit/"
        with logger.indent():
            logger.info(f"preview at {url}")
        headers = {
            "accept": "application/json",
            "Authorization": f"Token {self.readthedocs_token}",
        }
        endpoint = "https://readthedocs.org/api/v3"

        url = f"{endpoint}/projects/{readthedocs_project_name_slug}/"
        if real_run:
            response = requests.get(url, headers=headers)
            if response.status_code == 200:
                url = f"https://app.readthedocs.org/projects/{readthedocs_project_name_slug}/"
                logger.info(
                    f"Project already exists on readthedocs.org, "
                    f"please view it at: {url}"
                )
                return True
            elif response.status_code == 404:
                logger.info("Project does not exist on readthedocs.org, creating it...")
            else:
                raise_http_response_error(response)

        url = f"{endpoint}/projects/"
        data = {
            "name": self.readthedocs_project_name,
            "repository": {"url": self.github_repo_url, "type": "git"},
            "homepage": f"http://{readthedocs_project_name_slug}.readthedocs.io/",
            "programming_language": "py",
            "language": "en",
            "privacy_level": "public",
            "external_builds_privacy_level": "public",
            "versioning_scheme": "multiple_versions_with_translations",
            "tags": [],
        }
        if real_run:
            response = requests.post(
                url,
                headers=headers,
                json=data,
            )
            if response.status_code < 200 or response.status_code >= 300:
                raise_http_response_error(response)
        return real_run

    def setup_readthedocs_project(
        self: "PyWf",
        real_run: bool = True,
        verbose: bool = True,
    ):
        with logger.disabled(not verbose):
            return self._setup_readthedocs_project(
                real_run=real_run,
            )

    setup_readthedocs_project.__doc__ = _setup_readthedocs_project.__doc__



# -*- coding: utf-8 -*-

"""
Build Python source distribution.

Usage example:

.. code-block:: python

    from build_dist import build_dist_with_python_build, build_dist_with_poetry_build
"""

import typing as T
import os
import contextlib
import subprocess
from pathlib import Path

__version__ = "0.1.1"


@contextlib.contextmanager
def temp_cwd(path: T.Union[str, Path]):
    """
    Temporarily set the current directory to target path.
    """
    path = Path(path).absolute()
    if not path.is_dir():
        raise ValueError(f"{path} is not a dir!")

    cwd = os.getcwd()
    os.chdir(f"{path}")
    try:
        yield path
    finally:
        os.chdir(cwd)


def build_dist_with_python_build(
    dir_project_root: T.Union[str, Path],
    path_bin_python: T.Union[str, Path],
    log_func: T.Callable = print,
    real_run: bool = True,
    verbose: bool = True,
):
    """
    Build the source distribution with ``python-build``.

    :param dir_project_root: the root directory of your project, it should have
        a setup.py or pyproject.toml file
    :param path_bin_python: the path to python executable, usually the
        virtualenv python
    :param verbose: show verbose output or not

    Reference: https://pypa-build.readthedocs.io/en/latest/
    """
    dir_project_root = Path(dir_project_root).absolute()
    path_bin_python = Path(path_bin_python).absolute()
    with temp_cwd(dir_project_root):
        args = [
            f"{path_bin_python}",
            "-m",
            "build",
            "--sdist",
            "--wheel",
        ]
        log_func("run command: {}".format(" ".join(args)))
        if real_run:
            subprocess.run(args, check=True, capture_output=not verbose)


def build_dist_with_poetry_build(
    dir_project_root: T.Union[str, Path],
    path_bin_poetry: T.Union[str, Path],
    log_func: T.Callable = print,
    real_run: bool = True,
    verbose: bool = True,
):
    """
    Build the source distribution with ``poetry build`` command.

    :param dir_project_root: the root directory of your project, it should have
        a setup.py or pyproject.toml file
    :param path_bin_poetry: the path to poetry executable, could be simply "poetry"
    :param verbose: show verbose output or not

    Reference: https://python-poetry.org/docs/cli/#build
    """
    dir_project_root = Path(dir_project_root).absolute()
    path_bin_poetry = Path(path_bin_poetry)  # poetry could be a global command
    with temp_cwd(dir_project_root):
        args = [
            f"{path_bin_poetry}",
            "build",
        ]
        if verbose is False:
            args.append("--quiet")
        log_func("run command: {}".format(" ".join(args)))
        if real_run:
            subprocess.run(args, check=True)
